# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import subprocess
from textwrap import dedent

import testscenarios
import yaml as yaml_utils
from testtools.matchers import Contains, FileExists

from tests import integration


class VersionScriptBaseTestCase(integration.TestCase):
    def _set_version_script(self, snapcraft_yaml_file):
        with open(snapcraft_yaml_file) as f:
            snapcraft_yaml = yaml_utils.load(f)
        snapcraft_yaml["version-script"] = self.scripts[self.script]
        with open(snapcraft_yaml_file, "w") as f:
            yaml_utils.dump(snapcraft_yaml, stream=f)


class VersionScriptTestCase(testscenarios.WithScenarios, VersionScriptBaseTestCase):

    scripts = {
        "empty": "",  # this is 0.1
        "simple": "echo from-version-script",
        "exported-version": (
            "echo from-variable-$SNAPCRAFT_PROJECT_VERSION"
        ),  # this is 0.1
        "exported-version-and-grade": (
            "echo $SNAPCRAFT_PROJECT_VERSION-$SNAPCRAFT_PROJECT_GRADE"
        ),
        "multi-line": dedent(
            """\
            if [ "$SNAPCRAFT_PROJECT_GRADE" = "devel" ] ; then
                echo "development-build"
            else
                exit 1
            fi
            """
        ),
        "cat-file": "cat stage/version",
    }

    scenarios = [
        ("empty", dict(script="empty", expected_version="0.1")),
        ("simple", dict(script="simple", expected_version="from-version-script")),
        (
            "version",
            dict(script="exported-version", expected_version="from-variable-0.1"),
        ),
        (
            "version-and-grade",
            dict(script="exported-version-and-grade", expected_version="0.1-devel"),
        ),
        ("multi-line", dict(script="multi-line", expected_version="development-build")),
        ("cat-file", dict(script="cat-file", expected_version="test-build")),
    ]

    def test_version(self):
        self.copy_project_to_cwd("version-script")
        self._set_version_script(os.path.join("snap", "snapcraft.yaml"))

        self.run_snapcraft("snap")

        self.assertThat(
            "version-script-test_{}_{}.snap".format(
                self.expected_version, self.deb_arch
            ),
            FileExists(),
        )


class VersionScriptErrorsTestCase(
    testscenarios.WithScenarios, VersionScriptBaseTestCase
):

    scripts = {"empty-version": "echo", "exit-2": "exit 2"}

    scenarios = (
        (
            "no output",
            dict(script="empty-version", error="The version-script produced no output"),
        ),
        (
            "exit 2",
            dict(
                script="exit-2", error="The version-script failed to run (exit code 2)"
            ),
        ),
    )

    def test_no_output(self):
        self.copy_project_to_cwd("version-script")
        self._set_version_script(os.path.join("snap", "snapcraft.yaml"))

        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, ["snap"]
        )
        self.assertThat(exception.output, Contains(self.error))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from snapcraft.plugins.v1._python import errors


class TestErrorFormatting:

    scenarios = (
        (
            "PipListInvalidLegacyFormatError",
            {
                "exception_class": errors.PipListInvalidLegacyFormatError,
                "kwargs": {"output": "test-output"},
                "expected_message": (
                    "Failed to parse Python package list: "
                    "The returned output is not in the expected format:\n"
                    "test-output"
                ),
            },
        ),
    )

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The catkin plugin for ROS 1 parts.

    - catkin-packages:
      (list of strings)
      List of catkin packages to build. If not specified, all packages in the
      workspace will be built. If set to an empty list ([]), no packages will
      be built, which could be useful if you only want ROS debs in the snap.

    - catkin-packages-ignore:
      (list of strings)
      List of catkin packages to ignore (i.e. not build or install). If not
      specified or set to an empty list ([]), no packages will be ignored.

    - catkin-cmake-args:
      (list of strings)
      Arguments to pass to cmake projects.

    This plugin requires certain variables that are specified by the `ros1-noetic`
    extension. If you're not using the extension, set these in your `build-environment`:
      - ROS_DISTRO: "noetic"
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import _ros


class CatkinPlugin(_ros.RosPlugin):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "catkin-cmake-args": {
                    "type": "array",
                    "minItems": 0,
                    "items": {"type": "string"},
                    "default": [],
                },
                "catkin-packages": {
                    "type": "array",
                    "minItems": 0,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                },
                "catkin-packages-ignore": {
                    "type": "array",
                    "minItems": 0,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
            },
        }

    def get_build_packages(self) -> Set[str]:
        return super().get_build_packages() | {"ros-noetic-catkin"}

    def _get_build_commands(self) -> List[str]:
        cmd = [
            "catkin_make_isolated",
            "--install",
            "--merge",
            "--source-space",
            "$SNAPCRAFT_PART_SRC",
            "--build-space",
            "$SNAPCRAFT_PART_BUILD",
            "--install-space",
            "$SNAPCRAFT_PART_INSTALL/opt/ros/$ROS_DISTRO",
            "-j",
            "$SNAPCRAFT_PARALLEL_BUILD_COUNT",
        ]

        if self.options.catkin_packages:
            cmd.extend(["--pkg", *self.options.catkin_packages])

        if self.options.catkin_packages_ignore:
            cmd.extend(["--ignore-pkg", *self.options.catkin_packages_ignore])

        if self.options.catkin_cmake_args:
            cmd.extend(["--cmake-args", *self.options.catkin_cmake_args])

        return [" ".join(cmd)]

    def _get_workspace_activation_commands(self) -> List[str]:
        return [
            'state="$(set +o)"',
            "set +u",
            "_CATKIN_SETUP_DIR=/opt/ros/$ROS_DISTRO . /opt/ros/$ROS_DISTRO/setup.sh",
            'eval "$(state)"',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import glob
import os
import pathlib
from textwrap import dedent

import fixtures
from testtools.matchers import Contains, Equals, Not

import snapcraft
import tests
from snapcraft import file_utils
from snapcraft.internal import cache
from tests.unit.commands import CommandBaseTestCase


class SnapCacheBaseTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()

        self.deb_arch = snapcraft.ProjectOptions().deb_arch
        self.snap_path = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap.snap"
        )


class SnapCacheTestCase(SnapCacheBaseTestCase):
    def test_snap_cache(self):
        # cache snap
        snap_cache = cache.SnapCache(project_name="cache-test")
        cached_snap_path = snap_cache.cache(snap_filename=self.snap_path)

        expected_snap_path = os.path.join(
            snap_cache.snap_cache_root,
            "amd64",
            file_utils.calculate_sha3_384(self.snap_path),
        )

        self.assertThat(cached_snap_path, Equals(expected_snap_path))
        self.assertTrue(os.path.isfile(cached_snap_path))

    def test_snap_cache_get_latest(self):
        # Create snaps
        meta_dir = os.path.join(self.path, "prime", "meta")
        os.makedirs(meta_dir)
        with open(os.path.join(meta_dir, "snap.yaml"), "w") as f:
            print(
                dedent(
                    """\
                name: my-snap-name
                summary: test cached snap
                description: test cached snap
                confinement: devmode
                grade: devel
                version: '0.1'
                """
                ),
                file=f,
            )
        result = self.run_command(["pack", os.path.join(self.path, "prime")])
        self.assertThat(result.exit_code, Equals(0))
        snap_file = glob.glob("*0.1*.snap")[0]

        snap_cache = cache.SnapCache(project_name="my-snap-name")
        snap_cache.cache(snap_filename=snap_file)

        with open(os.path.join(meta_dir, "snap.yaml"), "w") as f:
            print(
                dedent(
                    """\
                name: my-snap-name
                summary: test cached snap
                description: test cached snap
                confinement: devmode
                grade: devel
                version: '0.2'
                """
                ),
                file=f,
            )
        result = self.run_command(["pack", os.path.join(self.path, "prime")])
        self.assertThat(result.exit_code, Equals(0))
        snap_file_latest = glob.glob("*0.2*.snap")[0]

        snap_cache.cache(snap_filename=snap_file_latest)
        latest_hash = file_utils.calculate_sha3_384(snap_file_latest)

        # get latest
        latest_snap = snap_cache.get(deb_arch="all")

        expected_snap_path = os.path.join(
            snap_cache.snap_cache_root, "all", latest_hash
        )

        self.assertThat(latest_snap, Equals(expected_snap_path))

    def test_snap_cache_get_by_hash(self):
        snap_cache = cache.SnapCache(project_name="my-snap-name")
        snap_cache.cache(snap_filename=self.snap_path)

        # get hash of snap
        snap_hash = file_utils.calculate_sha3_384(self.snap_path)

        # get snap by hash
        snap = snap_cache.get(deb_arch="amd64", snap_hash=snap_hash)

        self.assertThat(
            snap, Equals(os.path.join(snap_cache.snap_cache_root, "amd64", snap_hash))
        )


class SnapCachePruneTestCase(SnapCacheBaseTestCase):
    def test_prune_snap_cache(self):
        self.useFixture(
            fixtures.MockPatchObject(
                cache.SnapCache, "_get_snap_deb_arch", return_value="amd64"
            )
        )
        # Create snaps
        snap_file_1 = pathlib.Path("snap_01_amd64.snap")
        snap_file_2 = pathlib.Path("snap_02_amd64.snap")

        for snap_file in (snap_file_1, snap_file_2):
            with snap_file.open("w") as f:
                # Add whatever content
                print(snap_file.as_posix(), file=f)

        snap_cache = cache.SnapCache(project_name="snap")
        snap_file_1_path = snap_cache.cache(snap_filename=snap_file_1)
        _, snap_file_1_hash = os.path.split(snap_file_1_path)

        snap_file_2_path = snap_cache.cache(snap_filename=snap_file_2)
        snap_file_2_dir, snap_file_2_hash = os.path.split(snap_file_2_path)

        # confirm expected snap cached
        self.assertThat(len(os.listdir(snap_file_2_dir)), Equals(2))

        # prune
        pruned_files = snap_cache.prune(
            deb_arch=self.deb_arch, keep_hash=snap_file_2_hash
        )

        self.assertThat(len(pruned_files), Equals(1))
        self.assertThat(
            pruned_files,
            Contains(
                os.path.join(
                    snap_cache.snap_cache_root, self.deb_arch, snap_file_1_hash
                )
            ),
        )

        self.assertThat(
            pruned_files,
            Not(Contains(os.path.join(snap_cache.snap_cache_root, snap_file_2_hash))),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib

from snapcraft.internal.db import errors


def test_SnapcraftDatastoreVersionUnsupported():
    error = errors.SnapcraftDatastoreVersionUnsupported(
        path=pathlib.Path("/tmp/foo.yaml"), current_version=5, supported_version=4
    )

    assert (
        error.get_brief()
        == "This version of snapcraft does not support version 5 of the /tmp/foo.yaml datastore."
    )
    assert error.get_details() is None
    assert (
        error.get_resolution()
        == "Use 'snap revert' or 'snap refresh' to install previously used version of snapcraft (or newer)."
    )
    assert error.get_docs_url() is None
    assert error.get_reportable() is False
    assert error.get_exit_code() == 2



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from testtools.matchers import (
    Annotate,
    Contains,
    DirExists,
    Equals,
    FileExists,
    MatchesRegex,
    Not,
)

from tests import fixture_setup, integration, skip


class RpathTestCase(integration.TestCase):
    def test_origin_rpath(self):
        self.run_snapcraft("prime", "rpath-test")

        self.assertThat(os.path.join(self.prime_dir, "binary"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "lib", "libfoo.so"), FileExists())

        # Assert that the $ORIGIN rpath did not result in the library being
        # pulled in twice.
        self.assertThat(
            os.path.join(self.prime_dir, os.path.abspath(self.prime_dir).lstrip("/")),
            Not(DirExists()),
        )


class Libc6TestCase(integration.TestCase):
    def test_primed_libc6(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, base=None)
        snapcraft_yaml.update_part(
            "test-part", {"plugin": "nil", "stage-packages": ["libc6", "hello"]}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")

        bin_path = os.path.join(self.prime_dir, "usr", "bin", "hello")
        self.assertThat(bin_path, FileExists())

        # Verify that the correct interpreter that has been staged is used.
        interpreter = subprocess.check_output(
            [self.patchelf_command, "--print-interpreter", bin_path]
        ).decode()
        expected_interpreter = r"^/snap/test-snap/current/lib/.*"
        self.assertThat(interpreter, MatchesRegex(expected_interpreter))


class OriginRPATHTestCase(integration.TestCase):
    def test_origin(self):
        # We stage libc6 for this to work on non xenial
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, base=None)
        snapcraft_yaml.update_part(
            "test-part", {"plugin": "nil", "stage-packages": ["libc6", "python3"]}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")

        bin_path = os.path.join(self.prime_dir, "usr", "bin", "python3")
        self.assertThat(bin_path, FileExists())

        # Verify there aren't any duplicate rpath entries
        rpath = (
            subprocess.check_output([self.patchelf_command, "--print-rpath", bin_path])
            .decode()
            .strip()
        )
        seen = set()
        for i in rpath.split(":"):
            self.assertThat(
                seen, Annotate("A duplicate rpath was found!", Not(Contains(i)))
            )
            seen.add(i)

    @skip.skip_unless_codename("xenial", "we currently only have core")
    def test_origin_runpath_migrated_to_rpath(self):
        # We stage libc6 for this to work on non xenial
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, base=None, confinement="classic"
        )
        snapcraft_yaml.update_part(
            "test-part",
            {
                "plugin": "nil",
                "build-packages": ["patchelf"],
                "stage-packages": ["hello"],
                # Without --force-rpath this is essentially a RUNPATH
                "override-build": (
                    "patchelf --set-rpath '$ORIGIN/foobar' "
                    "$SNAPCRAFT_PART_INSTALL/usr/bin/hello"
                ),
            },
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")

        bin_path = os.path.join(self.prime_dir, "usr", "bin", "hello")
        self.assertThat(bin_path, FileExists())

        # Verify there aren't any duplicate rpath entries
        readelf_d = (
            subprocess.check_output(["readelf", "-d", bin_path]).decode().strip()
        )
        lines_with_runpath = [l for l in readelf_d.splitlines() if "RUNPATH" in l]
        lines_with_rpath = [l for l in readelf_d.splitlines() if "RPATH" in l]
        self.assertThat(
            len(lines_with_runpath),
            Annotate(
                "Expected no RUNPATH entries but found {}".format(lines_with_runpath),
                Equals(0),
            ),
        )
        self.assertThat(
            len(lines_with_rpath),
            Annotate(
                "Expected one RPATH entry but found {}".format(lines_with_rpath),
                Equals(1),
            ),
        )


class ExecStackTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()
        if self.deb_arch == "arm64":
            self.skipTest("architecture is not supported by execstack")

    def _setup_project(self, keep_execstack: bool):
        if keep_execstack:
            attributes = ["keep-execstack"]
        else:
            attributes = []

        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, base=None)
        snapcraft_yaml.update_part(
            "test-part",
            {
                "plugin": "nil",
                "build-attributes": attributes,
                "override-build": (
                    "execstack --set-execstack $SNAPCRAFT_PART_INSTALL/usr/bin/hello"
                ),
                "prime": ["usr/bin/hello"],
                "build-packages": ["execstack"],
                "stage-packages": ["hello"],
            },
        )
        self.useFixture(snapcraft_yaml)

    def _assert_execstack(self, is_set: bool):
        stage_bin_path = os.path.join("stage", "usr", "bin", "hello")
        stage_bin_query = (
            subprocess.check_output([self.execstack_command, "--query", stage_bin_path])
            .decode()
            .strip()
        )
        # 'X' means the executable stack is required. See `man 8 execstack`.
        self.assertThat(stage_bin_query, Equals("X {}".format(stage_bin_path)))

        prime_bin_path = os.path.join("prime", "usr", "bin", "hello")
        prime_bin_query = (
            subprocess.check_output([self.execstack_command, "--query", prime_bin_path])
            .decode()
            .strip()
        )
        # '-' means the executable stack is not required.
        # See `man 8 execstack`.
        if is_set:
            expected_value = "X"
        else:
            expected_value = "-"
        self.assertThat(
            prime_bin_query, Equals("{} {}".format(expected_value, prime_bin_path))
        )

    def test_keep_execstack(self):
        self._setup_project(keep_execstack=True)
        self.run_snapcraft("prime")

        self._assert_execstack(is_set=True)

    def test_clear_execstack(self):
        self._setup_project(keep_execstack=False)
        self.run_snapcraft("prime")

        self._assert_execstack(is_set=False)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import filecmp
import os
import shutil
import subprocess
import sys
import tempfile

import apt
import fixtures
import testscenarios
import yaml as yaml_utils
from testtools.matchers import Contains, Equals

from tests import fixture_setup, integration, os_release
from tests.integration import repo


class AssetRecordingBaseTestCase(integration.TestCase):
    """Test that the prime step records an annotated manifest.yaml

    The annotated file will be in prime/snap/manifest.yaml.

    """

    def setUp(self):
        super().setUp()
        # The combination of snapd, lxd and armhf does not currently work.
        if os.environ.get("ADT_TEST") and self.deb_arch == "armhf":
            self.skipTest("The autopkgtest armhf runners can't install snaps")

        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_INFO", "1"))

        try:
            subprocess.check_output(["snap", "list", "review-tools"])
        except subprocess.CalledProcessError:
            subprocess.check_call(["sudo", "snap", "install", "review-tools", "--edge"])

    def assert_review_passes(self, snap_file: str) -> None:
        # See https://forum.snapcraft.io/t/7339 for more info on why review verification
        # is skipped on non xenial environments.
        if os_release.get_version_codename() != "xenial":
            return

        # review-tools do not really have access to tmp, let's assume it can look
        # in its own snap directory and that that does not change as we cannot
        # query what the data store is for a given snap.
        review_tools_common_dir = os.path.expanduser(
            os.path.join("~", "snap", "review-tools", "common")
        )
        os.makedirs(review_tools_common_dir, exist_ok=True)
        with tempfile.NamedTemporaryFile(dir=review_tools_common_dir) as temp_snap_file:
            shutil.copyfile(snap_file, temp_snap_file.name)
            try:
                subprocess.check_output(
                    ["review-tools.snap-review", temp_snap_file.name]
                )
            except subprocess.CalledProcessError as call_error:
                self.fail(
                    "{!r} does not pass the review:\n{}".format(
                        snap_file, call_error.stdout.decode()
                    )
                )


class SnapcraftYamlRecordingTestCase(AssetRecordingBaseTestCase):
    def test_prime_records_snapcraft_yaml_copy(self):
        self.run_snapcraft("prime", project_dir="basic")
        source_snapcraft_yaml_path = os.path.join("snap", "snapcraft.yaml")
        recorded_snapcraft_yaml_path = os.path.join(
            self.prime_dir, "snap", "snapcraft.yaml"
        )
        self.assertTrue(
            filecmp.cmp(
                source_snapcraft_yaml_path, recorded_snapcraft_yaml_path, shallow=False
            )
        )


class ManifestRecordingTestCase(AssetRecordingBaseTestCase):
    def test_prime_records_uname(self):
        self.run_snapcraft(["snap", "--output", "basic.snap"], project_dir="basic")

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        expected_uname = (
            subprocess.check_output(["uname", "-srvmpio"])
            .decode(sys.getfilesystemencoding())
            .strip()
        )
        self.assertThat(
            recorded_yaml["parts"]["dummy-part"]["uname"], Equals(expected_uname)
        )
        self.assert_review_passes("basic.snap")

    def test_prime_records_installed_packages(self):
        self.run_snapcraft(["snap", "--output", "basic.snap"], project_dir="basic")

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        with apt.Cache() as apt_cache:
            expected_package = "python3={}".format(
                apt_cache["python3"].installed.version
            )
        self.assertThat(
            recorded_yaml["parts"]["dummy-part"]["installed-packages"],
            Contains(expected_package),
        )
        self.assert_review_passes("basic.snap")

    def test_prime_records_installed_snaps(self):
        subprocess.check_call(["sudo", "snap", "install", "core"])
        self.run_snapcraft(["snap", "--output", "basic.snap"], project_dir="basic")

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        expected_package = "core={}".format(
            repo.get_local_snap_info("core")["revision"]
        )
        self.assertThat(
            recorded_yaml["parts"]["dummy-part"]["installed-snaps"],
            Contains(expected_package),
        )
        self.assert_review_passes("basic.snap")

    def test_prime_with_architectures(self):
        """Test the recorded manifest for a basic snap

        This snap doesn't have stage or build packages and is declared that it
        works on all architectures.
        """
        self.run_snapcraft(["snap", "--output", "basic.snap"], project_dir="basic")

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(recorded_yaml["architectures"], Equals(["all"]))
        self.assert_review_passes("basic.snap")

    def test_prime_without_architectures_records_current_arch(self):
        """Test the recorded manifest for a basic snap

        This snap doesn't have stage or build packages and it is not declared
        that it works on all architectures, which makes it specific to the
        current architecture.
        """
        self.run_snapcraft(
            ["snap", "--output", "basic-without-arch.snap"],
            project_dir="basic-without-arch",
        )

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(recorded_yaml["architectures"], Equals([self.deb_arch]))
        self.assert_review_passes("basic-without-arch.snap")

    def test_prime_records_build_snaps(self):
        self.useFixture(fixture_setup.WithoutSnapInstalled("hello"))
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "test-part", {"plugin": "nil", "build-snaps": ["hello"]}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")

        expected_revision = repo.get_local_snap_info("hello")["revision"]
        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        # spread tests might leave bases behind which would affect this check
        self.assertThat(
            recorded_yaml["build-snaps"], Contains("hello={}".format(expected_revision))
        )


class ManifestRecordingBuildPackagesTestCase(
    testscenarios.WithScenarios, AssetRecordingBaseTestCase
):

    scenarios = [
        (snap, {"snap": snap})
        for snap in [
            "build-package-global",
            "build-packages-missing-dependency",
            "build-packages-without-dependencies",
        ]
    ]

    def test_prime_with_build_packages(self):
        """Test the recorded manifest for a snap with build packages

        This snap declares one global build package that has undeclared
        dependencies.

        """
        expected_packages = ["haskell-doc", "haskell98-tutorial", "haskell98-report"]
        self.addCleanup(
            subprocess.call, ["sudo", "apt", "remove", "-y"] + expected_packages
        )

        self.run_snapcraft("prime", self.snap)

        expected_packages_with_version = [
            "{}={}".format(
                package,
                integration.get_package_version(
                    package, self.distro_series, self.deb_arch
                ),
            )
            for package in expected_packages
        ]

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(
            recorded_yaml["build-packages"], Equals(expected_packages_with_version)
        )


class ManifestRecordingStagePackagesTestCase(AssetRecordingBaseTestCase):
    def test_prime_records_packages_version(self):
        """Test the recorded manifest for a snap with packages

        This snap declares all the packages that it requires, there are
        no additional dependencies. The packages specify their version.
        """
        expected_packages = ["haskell-doc", "haskell98-tutorial", "haskell98-report"]
        self.copy_project_to_cwd("stage-packages-without-dependencies")
        part_name = "part-with-stage-packages"
        for package in expected_packages:
            self.set_stage_package_version(
                os.path.join("snap", "snapcraft.yaml"), part_name, package
            )

        self.run_snapcraft("prime")

        with open(os.path.join("snap", "snapcraft.yaml")) as source_yaml_file:
            source_yaml = yaml_utils.load(source_yaml_file)

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(
            recorded_yaml["parts"][part_name]["stage-packages"],
            Equals(source_yaml["parts"][part_name]["stage-packages"]),
        )

    def test_prime_without_packages_version(self):
        """Test the recorded manifest for a snap with packages

        This snap declares all the packages that it requires, there are
        no additional dependencies. The packages don't specify their
        version.
        """
        self.run_snapcraft("prime", project_dir="stage-packages-without-dependencies")

        with open(os.path.join("snap", "snapcraft.yaml")) as source_yaml_file:
            source_yaml = yaml_utils.load(source_yaml_file)
        part_name = "part-with-stage-packages"
        expected_packages = [
            "{}={}".format(
                package,
                integration.get_package_version(
                    package, self.distro_series, self.deb_arch
                ),
            )
            for package in source_yaml["parts"][part_name]["stage-packages"]
        ]

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(
            recorded_yaml["parts"][part_name]["stage-packages"],
            Equals(expected_packages),
        )

    def test_prime_with_packages_missing_dependency(self):
        """Test the recorded manifest for a snap with packages

        This snap declares one package that has undeclared dependencies.
        """
        self.copy_project_to_cwd("stage-packages-missing-dependency")
        part_name = "part-with-stage-packages"
        self.set_stage_package_version(
            os.path.join("snap", "snapcraft.yaml"), part_name, package="haskell-doc"
        )
        self.run_snapcraft("prime")

        expected_packages = [
            "{}={}".format(
                package,
                integration.get_package_version(
                    package, self.distro_series, self.deb_arch
                ),
            )
            for package in ["haskell-doc", "haskell98-tutorial", "haskell98-report"]
        ]

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(
            recorded_yaml["parts"][part_name]["stage-packages"],
            Equals(expected_packages),
        )


class ManifestRecordingBzrSourceTestCase(
    integration.BzrSourceBaseTestCase, AssetRecordingBaseTestCase
):
    def test_prime_with_bzr_source(self):
        self.copy_project_to_cwd("bzr-head")

        self.init_source_control()
        self.commit('"test-commit"', unchanged=True)

        self.run_snapcraft(["snap", "--output", "bzr.snap"])

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        commit = self.get_revno()
        self.assertThat(recorded_yaml["parts"]["bzr"]["source-commit"], Equals(commit))
        self.assert_review_passes("bzr.snap")


class ManifestRecordingGitSourceTestCase(
    integration.GitSourceBaseTestCase, AssetRecordingBaseTestCase
):
    def test_prime_with_git_source(self):
        self.copy_project_to_cwd("git-head")

        self.init_source_control()
        self.commit('"test-commit"', allow_empty=True)

        self.run_snapcraft(["snap", "--output", "git.snap"])

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        commit = self.get_revno()
        self.assertThat(recorded_yaml["parts"]["git"]["source-commit"], Equals(commit))
        self.assert_review_passes("git.snap")


class ManifestRecordingHgSourceTestCase(
    integration.HgSourceBaseTestCase, AssetRecordingBaseTestCase
):
    def test_prime_with_hg_source(self):
        self.copy_project_to_cwd("hg-head")

        self.init_source_control()
        open("1", "w").close()
        self.commit("1", "1")

        self.run_snapcraft(["snap", "--output", "hg.snap"])

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        commit = self.get_id()
        self.assertThat(
            recorded_yaml["parts"]["mercurial"]["source-commit"], Equals(commit)
        )
        self.assert_review_passes("hg.snap")


class ManifestRecordingSubversionSourceTestCase(
    integration.SubversionSourceBaseTestCase, AssetRecordingBaseTestCase
):
    def test_prime_with_subversion_source(self):
        self.copy_project_to_cwd("svn-pull")

        self.init_source_control()
        self.checkout("file:///{}".format(os.path.join(self.path, "repo")), "local")

        open(os.path.join("local", "file"), "w").close()
        self.add("file", cwd="local")
        self.commit("test", cwd="local/")
        self.update(cwd="local/")
        subprocess.check_call(["rm", "-rf", "local/"], stdout=subprocess.DEVNULL)

        self.run_snapcraft(["snap", "--output", "svn.snap"])

        recorded_yaml_path = os.path.join(self.prime_dir, "snap", "manifest.yaml")
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml_utils.load(recorded_yaml_file)

        self.assertThat(recorded_yaml["parts"]["svn"]["source-commit"], Equals("1"))
        self.assert_review_passes("svn.snap")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap

from testtools.matchers import Contains, Equals

import snapcraft.storeapi.errors

from . import StoreCommandsBaseTestCase

account_info_data = {"snaps": {"16": {"core": {"snap-id": "good"}}}}


class GatedCommandTestCase(StoreCommandsBaseTestCase):
    def test_gated_unknown_snap(self):
        self.client.login(email="dummy", password="test correct password")

        raised = self.assertRaises(
            snapcraft.storeapi.errors.SnapNotFoundError,
            self.run_command,
            ["gated", "notfound"],
        )

        self.assertThat(str(raised), Equals("Snap 'notfound' was not found."))

    def test_gated_success(self):
        self.client.login(email="dummy", password="test correct password")

        result = self.run_command(["gated", "core"])

        self.assertThat(result.exit_code, Equals(0))
        expected_output = textwrap.dedent(
            """\
            Name      Revision  Required    Approved
            snap-1           3  True        2016-09-19T21:07:27Z
            snap-2           5  False       2016-09-19T21:07:27Z
            snap-3           -  True        2016-09-19T21:07:27Z"""
        )
        self.assertThat(result.output, Contains(expected_output))

    def test_gated_no_validations(self):
        self.client.login(email="dummy", password="test correct password")

        result = self.run_command(["gated", "test-snap-with-no-validations"])

        self.assertThat(result.exit_code, Equals(0))
        expected_output = (
            "There are no validations for snap 'test-snap-with-no-validations'\n"
        )
        self.assertThat(result.output, Contains(expected_output))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os
import sys
import zipfile

from testtools.matchers import Equals

from snapcraft.internal.sources import errors
from snapcraft.internal.sources._checksum import verify_checksum
from tests import unit

if sys.version_info < (3, 6):
    import sha3  # noqa


class TestChecksum(unit.TestCase):
    def setUp(self):
        super().setUp()

    def test_invalid_checksum(self):
        # Create a file for testing
        os.makedirs(os.path.join("src"))
        dummy_file = os.path.join("src", "test")
        open(dummy_file, "w").close()

        self.assertRaises(AttributeError, verify_checksum, "456/abcde", dummy_file)

    def test_correct_checksum(self):
        # Create zip file for testing
        os.makedirs(os.path.join("src"))
        file_to_zip = os.path.join("src", "test.txt")
        open(file_to_zip, "w").close()
        zip_file = zipfile.ZipFile(os.path.join("src", "test.zip"), "w")
        zip_file.write(file_to_zip)
        zip_file.close()

        calculated_checksum = hashlib.new(
            "md5", open(os.path.join("src", "test.zip"), "rb").read()
        )
        calculated_checksum = calculated_checksum.hexdigest()

        verify_checksum("md5/" + calculated_checksum, "src/test.zip")

    def test_incorrect_checksum(self):
        # Create zip file for testing
        os.makedirs(os.path.join("src"))
        file_to_zip = os.path.join("src", "test.txt")
        open(file_to_zip, "w").close()
        zip_file = zipfile.ZipFile(os.path.join("src", "test.zip"), "w")
        zip_file.write(file_to_zip)
        zip_file.close()

        incorrect_checksum = "fe049cfba688aa1af88bc78191d7f904"

        calculated_checksum = hashlib.new(
            "md5", open(os.path.join("src", "test.zip"), "rb").read()
        )
        calculated_checksum = calculated_checksum.hexdigest()

        raised = self.assertRaises(
            errors.DigestDoesNotMatchError,
            verify_checksum,
            "md5/" + incorrect_checksum,
            "src/test.zip",
        )

        self.assertThat(raised.expected, Equals(incorrect_checksum))
        self.assertThat(raised.calculated, Equals(calculated_checksum))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.repo.deb_package import DebPackage


def test_basic():
    package = DebPackage(name="foo", arch="arch", version="1.0")

    assert package.name == "foo"
    assert package.arch == "arch"
    assert package.version == "1.0"


def test_parse_simple():
    assert DebPackage.from_unparsed("foo") == DebPackage(
        name="foo", arch=None, version=None
    )


def test_parse_arch():
    assert DebPackage.from_unparsed("foo:arch") == DebPackage(
        name="foo", arch="arch", version=None
    )


def test_parse_arch_and_version():
    assert DebPackage.from_unparsed("foo:arch=4.5") == DebPackage(
        name="foo", arch="arch", version="4.5"
    )


def test_parse_version():
    assert DebPackage.from_unparsed("foo=4.5") == DebPackage(
        name="foo", arch=None, version="4.5"
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import json
from snapcraft.internal.errors import details_from_command_error
import subprocess
import tempfile
from datetime import datetime
from textwrap import dedent
from typing import Any, Dict, List

import click
from tabulate import tabulate

import snapcraft
from snapcraft._store import StoreClientCLI
from snapcraft import storeapi, yaml_utils
from snapcraft.storeapi import assertions

from . import echo

_COLLABORATION_HEADER = dedent(
    """\
    # Change which developers may push or release snaps on the publisher's behalf.
    #
    # Sample entry:
    #
    # developers:
    #   - developer-id: "dev-one"      # Which developer
    #     since: "2017-02-10 08:35:00" # When contributions started
    #     until: "2018-02-10 08:35:00" # When contributions ceased (optional)
    #
    # All timestamps are UTC, and the "now" special string will be replaced by
    # the current time. Do not remove entries or use an until time in the past
    # unless you want existing snaps provided by the developer to stop working."""
)  # noqa


_VALIDATIONS_SETS_SNAPS_TEMPLATE = dedent(
    """\
    snaps:
    #  - name: <name>  # The name of the snap.
    #    id:   <id>    # The ID of the snap. Optional, defaults to the current ID for
                       # the provided name.
    #    presence: [required|optional|invalid]  # Optional, defaults to required.
    #    revision: <n> # The revision of the snap. Optional.
"""
)

_VALIDATION_SETS_TEMPLATE = dedent(
    """\
    account-id: {account_id}
    name: {set_name}
    sequence: {sequence}
    # The revision for this validation set
    # revision: {revision}
    {snaps}
    """
)


@click.group()
def assertionscli():
    """Store assertion commands"""


@assertionscli.command("list-keys")
def list_keys():
    """List the keys available to sign assertions.

    This command has an alias of `keys`.
    """
    snapcraft.list_keys()


@assertionscli.command("create-key")
@click.argument("key-name", metavar="<key-name>", required=False)
def create_key(key_name: str) -> None:
    """Create a key to sign assertions."""
    snapcraft.create_key(key_name)


@assertionscli.command("register-key")
@click.argument("key-name", metavar="<key-name>", required=False)
@click.option(
    "--experimental-login",
    is_flag=True,
    help="*EXPERIMENTAL* Enables login through candid.",
    envvar="SNAPCRAFT_LOGIN",
)
def register_key(key_name: str, experimental_login: bool) -> None:
    """Register a key with the store to sign assertions."""
    snapcraft.register_key(key_name, use_candid=experimental_login)


@assertionscli.command("sign-build")
@click.option("--key-name", metavar="<key-name>")
@click.argument(
    "snap-file",
    metavar="<snap-file>",
    type=click.Path(exists=True, readable=True, resolve_path=True, dir_okay=False),
)
@click.option(
    "--local", is_flag=True, help="Do not upload the generated assertion to the store"
)
def sign_build(snap_file: str, key_name: str, local: bool) -> None:
    """Sign a built snap file and assert it using the developer's key."""
    snapcraft.sign_build(snap_file, key_name=key_name, local=local)


@assertionscli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("validations", metavar="<validation>...", nargs=-1, required=True)
@click.option("--key-name", metavar="<key-name>")
@click.option("--revoke/--no-revoke", default=False)
def validate(snap_name: str, validations: list, key_name: str, revoke: bool) -> None:
    """Validate a gated snap.

    Each validation can be presented with either syntax:

    -  <snap-name>=<revision>
    -  <snap-id>=<revision>
    """
    snapcraft.validate(snap_name, validations, revoke=revoke, key=key_name)


@assertionscli.command()
@click.argument("snap-name", metavar="<snap-name>")
def gated(snap_name: str) -> None:
    """Get the list of snaps and revisions gating a snap."""
    snapcraft.gated(snap_name)


@assertionscli.command("edit-collaborators")
@click.argument("snap-name", metavar="<snap-name>")
@click.option("--key-name", metavar="<key-name>")
def edit_collaborators(snap_name, key_name):
    """Edit the list of collaborators for <snap-name>.

    This command has an alias of `collaborators`.
    """
    dev_assertion = assertions.DeveloperAssertion(
        snap_name=snap_name, signing_key=key_name
    )
    developers = dev_assertion.get_developers()
    updated_developers = _update_developers(developers)
    new_dev_assertion = dev_assertion.new_assertion(developers=updated_developers)
    if new_dev_assertion.is_equal(dev_assertion):
        echo.warning("Aborting due to unchanged collaborators list.")
        return

    try:
        new_dev_assertion.push()
    except storeapi.errors.StoreValidationError as store_error:
        if store_error.error_list[0]["code"] != "revoked-uploads":
            raise store_error
        click.echo(
            "This will revoke the following collaborators: {!r}".format(
                " ".join(store_error.error_list[0]["extra"])
            )
        )
        if echo.confirm("Are you sure you want to continue?"):
            new_dev_assertion.push(force=True)
        else:
            echo.warning("The collaborators for this snap have not been altered.")


@assertionscli.command("list-validation-sets")
@click.option(
    "--name",
    metavar="<name>",
    help="Only show results for a given Validation Set name.",
)
@click.option("--sequence", metavar="<sequence>", help="Sequences to show.")
def list_validation_sets(name, sequence):
    """Get the list of validation sets.

    The sequence option can be a sequence number or a keyword.

    \b
    Examples:
        snapcraft list-validation-sets
        snapcraft list-validation-sets --sequence all
        snapcraft list-validation-sets --name my-set --sequence 1

    Refer to https://snapcraft.io/docs/validation-sets for further information
    on Validation Sets.
    """
    store_client = StoreClientCLI()
    asserted_validation_sets = store_client.get_validation_sets(
        name=name, sequence=sequence,
    )

    if not asserted_validation_sets.assertions and (name or sequence):
        echo.warning("No validation sets found for the requested name or sequence.")
    elif not asserted_validation_sets.assertions:
        echo.warning("No validation sets found for this account.")
    else:
        headers = ["Account-ID", "Name", "Sequence", "Revision", "When"]
        assertions = list()
        for assertion in asserted_validation_sets.assertions:
            assertions.append(
                [
                    assertion.account_id,
                    assertion.name,
                    assertion.sequence,
                    assertion.revision,
                    datetime.strptime(
                        assertion.timestamp, "%Y-%m-%dT%H:%M:%SZ"
                    ).strftime("%Y-%m-%d"),
                ]
            )

        click.echo(
            tabulate(assertions, numalign="left", headers=headers, tablefmt="plain")
        )


@assertionscli.command("edit-validation-sets")
@click.argument("account-id", metavar="<account-id>")
@click.argument("set-name", metavar="<set-name>")
@click.argument("sequence", metavar="<sequence>", type=int)
@click.option("--key-name", metavar="<key-name>")
def edit_validation_sets(account_id: str, set_name: str, sequence: int, key_name: str):
    """Edit the list of validations for <set-name>.

    Refer to https://snapcraft.io/docs/validation-sets for further information
    on Validation Sets.
    """
    store_client = StoreClientCLI()

    asserted_validation_sets = store_client.get_validation_sets(
        name=set_name, sequence=str(sequence)
    )

    try:
        # assertions should only have one item since a specific
        # sequence was requested.
        revision = asserted_validation_sets.assertions[0].revision
        snaps = yaml_utils.dump(
            {
                "snaps": [
                    s.marshal() for s in asserted_validation_sets.assertions[0].snaps
                ]
            }
        )
    except IndexError:
        # If there is no assertion for a given sequence, the store API
        # will return an empty list.
        revision = "0"
        snaps = _VALIDATIONS_SETS_SNAPS_TEMPLATE

    unverified_validation_sets = _VALIDATION_SETS_TEMPLATE.format(
        account_id=account_id,
        set_name=set_name,
        sequence=sequence,
        revision=revision,
        snaps=snaps,
    )

    edited_validation_sets = _edit_validation_sets(unverified_validation_sets)
    if edited_validation_sets == yaml_utils.load(unverified_validation_sets):
        echo.warning("No changes made.")
    else:
        build_assertion = store_client.post_validation_sets_build_assertion(
            validation_sets=edited_validation_sets
        )
        signed_validation_sets = _sign_assertion(
            build_assertion.marshal(), key_name=key_name
        )
        store_client.post_validation_sets(signed_validation_sets=signed_validation_sets)


def _edit_validation_sets(validation_sets: str) -> Dict[str, Any]:
    """Spawn an editor to modify the validation-sets."""
    editor_cmd = os.getenv("EDITOR", "vi")

    with tempfile.NamedTemporaryFile() as ft:
        ft.close()
        with open(ft.name, "w") as fw:
            print(validation_sets, file=fw)
        subprocess.run([editor_cmd, ft.name], check=True)
        with open(ft.name, "r") as fr:
            edited_validation_sets = yaml_utils.load(fr)

    return edited_validation_sets


def _sign_assertion(assertion: Dict[str, Any], *, key_name: str) -> bytes:
    cmdline = ["snap", "sign"]
    if key_name:
        cmdline += ["-k", key_name]
    snap_sign = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    signed_assertion, _ = snap_sign.communicate(input=json.dumps(assertion).encode())
    if snap_sign.returncode != 0:
        echo.exit_error(
            brief="Failed to sign assertion.",
            details=details_from_command_error(
                cmd=cmdline, returncode=snap_sign.returncode
            ),
            exit_code=snap_sign.returncode,
        )

    return signed_assertion


def _update_developers(developers: List[Dict[str, str]]) -> List[Dict[str, str]]:
    edit_friendly_developers = _reformat_time_for_editing(developers)
    updated_developers = _edit_developers(edit_friendly_developers)
    return _reformat_time_for_assertion(updated_developers)


def _edit_developers(developers: List[Dict[str, str]]) -> List[Dict[str, str]]:
    """Spawn an editor to modify the snap-developer assertion for a snap."""
    editor_cmd = os.getenv("EDITOR", "vi")

    developer_wrapper = {"developers": developers}

    with tempfile.NamedTemporaryFile() as ft:
        ft.close()
        with open(ft.name, "w") as fw:
            print(_COLLABORATION_HEADER, file=fw)
            yaml_utils.dump(developer_wrapper, stream=fw)
        subprocess.check_call([editor_cmd, ft.name])
        with open(ft.name, "r") as fr:
            developers = yaml_utils.load(fr).get("developers")
    return developers


def _reformat_time_for_editing(
    developers: List[Dict[str, str]], time_format: str = "%Y-%m-%d %H:%M:%S"
) -> List[Dict[str, str]]:
    reformatted_developers = []
    for developer in developers:
        developer_it = {"developer-id": developer["developer-id"]}
        for range in ["since", "until"]:
            if range in developer:
                date = datetime.strptime(developer[range], "%Y-%m-%dT%H:%M:%S.%fZ")
                developer_it[range] = datetime.strftime(date, time_format)
        reformatted_developers.append(developer_it)
    return reformatted_developers


def _reformat_time_for_assertion(
    developers: List[Dict[str, str]]
) -> List[Dict[str, str]]:
    reformatted_developers = []
    for developer in developers:
        developer_it = {"developer-id": developer["developer-id"]}
        for range_ in ["since", "until"]:
            if range_ in developer:
                if developer[range_] == "now":
                    date = datetime.now()
                else:
                    date = datetime.strptime(developer[range_], "%Y-%m-%d %H:%M:%S")
                # We don't care about microseconds because we cannot edit
                # later so we set that to 0.
                developer_it[range_] = date.strftime("%Y-%m-%dT%H:%M:%S.000000Z")
        reformatted_developers.append(developer_it)
    return reformatted_developers



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import json
import os
import pathlib
import tarfile
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Equals, FileExists, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import nodejs
from tests import fixture_setup, unit

from . import PluginsV1BaseTestCase


class NodePluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            source = "."
            nodejs_version = nodejs._NODEJS_VERSION
            nodejs_package_manager = "npm"
            nodejs_yarn_version = ""
            source = "."

        self.options = Options()

        # always have a package.json stub under source
        open("package.json", "w").close()

        patcher = mock.patch("snapcraft.internal.common.run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.internal.common.run_output")
        self.run_output_mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.run_output_mock.return_value = '{"dependencies": []}'

        patcher = mock.patch("snapcraft.sources.Tar")
        self.tar_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.nodejs_url = nodejs.get_nodejs_release(
            nodejs._NODEJS_VERSION, self.project.deb_arch
        )

        self.useFixture(fixture_setup.CleanEnvironment())

    def create_assets(
        self,
        plugin,
        package_name="test-nodejs",
        single_bin=False,
        skip_package_json=False,
    ):
        for directory in (plugin.sourcedir, plugin.builddir):
            os.makedirs(directory)
            if not skip_package_json:
                with open(os.path.join(directory, "package.json"), "w") as json_file:
                    json.dump(
                        dict(
                            name=package_name, bin=dict(run="index.js"), version="1.0"
                        ),
                        json_file,
                    )
            package_name = package_name.lstrip("@").replace("/", "-")
            packed_tar_file = os.path.join(directory, "{}-1.0.tgz".format(package_name))
            open(os.path.join(directory, "index.js"), "w").close()
        tarfile.TarFile(packed_tar_file, "w").close()

        def provision(directory, **kwargs):
            if directory == os.path.join(plugin.builddir, "package"):
                package_dir = os.path.join(plugin.builddir, "package")
                open(os.path.join(package_dir, "index.js"), "w").close()

        self.tar_mock().provision.side_effect = provision

        # Create a fake node bin
        os.makedirs(os.path.join(plugin._npm_dir, "bin"))
        open(os.path.join(plugin._npm_dir, "bin", "node"), "w").close()


class NodejsPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = nodejs.NodePlugin.schema()
        properties = schema["properties"]

        # Check nodejs-version
        self.assertTrue(
            "nodejs-version" in properties,
            'Expected "nodejs-version" to be included in ' "properties",
        )
        node_channel = properties["nodejs-version"]

        self.assertTrue(
            "type" in node_channel, 'Expected "type" to be included in "nodejs-version"'
        )
        self.assertThat(
            node_channel["type"],
            Equals("string"),
            'Expected "nodejs-version" "type" to be '
            '"string", but it was '
            '"{}"'.format(node_channel["type"]),
        )
        self.assertTrue(
            "default" in node_channel,
            'Expected "default" to be included in "nodejs-version"',
        )
        self.assertThat(
            node_channel["default"],
            Equals(nodejs._NODEJS_VERSION),
            'Expected "nodejs-version" "default" to be '
            '"{}", but it was '
            '"{}"'.format(nodejs._NODEJS_VERSION, node_channel["default"]),
        )

        # Check nodejs-yarn-version
        self.assertTrue(
            "nodejs-yarn-version" in properties,
            'Expected "nodejs-yarn-version" to be included in ' "properties",
        )
        node_channel = properties["nodejs-yarn-version"]

        self.assertTrue(
            "type" in node_channel,
            'Expected "type" to be included in "nodejs-yarn-version"',
        )
        self.assertThat(
            node_channel["type"],
            Equals("string"),
            'Expected "nodejs-yarn-version" "type" to be '
            '"string", but it was '
            '"{}"'.format(node_channel["type"]),
        )
        self.assertTrue(
            "default" in node_channel,
            'Expected "default" to be included in "nodejs-yarn-version"',
        )
        self.assertThat(
            node_channel["default"],
            Equals(""),
            'Expected "nodejs-yarn-version" "default" to be '
            '"", but it was '
            '"{}"'.format(node_channel["default"]),
        )

        # Check nodejs-package-manager
        self.assertTrue(
            "nodejs-package-manager" in properties,
            'Expected "nodejs-package-manager" to be included in ' "properties",
        )
        nodejs_package_manager = properties["nodejs-package-manager"]

        self.assertTrue(
            "type" in nodejs_package_manager,
            'Expected "type" to be included in "nodejs-package-manager"',
        )
        self.assertThat(
            nodejs_package_manager["type"],
            Equals("string"),
            'Expected "nodejs-package-manager" "type" to be '
            '"string", but it was '
            '"{}"'.format(nodejs_package_manager["type"]),
        )
        self.assertTrue(
            "default" in nodejs_package_manager,
            'Expected "default" to be included in "nodejs-package-manager"',
        )
        self.assertThat(
            nodejs_package_manager["default"],
            Equals("yarn"),
            'Expected "nodejs-package-manager" "default" to be '
            '"yarn", but it was '
            '"{}"'.format(nodejs_package_manager["default"]),
        )

    def test_get_pull_properties(self):
        expected_pull_properties = [
            "nodejs-version",
            "nodejs-package-manager",
            "nodejs-yarn-version",
        ]
        resulting_pull_properties = nodejs.NodePlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = []
        resulting_build_properties = nodejs.NodePlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class NodePluginTest(NodePluginBaseTest):
    def get_npm_cmd(self, plugin):
        return os.path.join(plugin._npm_dir, "bin", "npm")

    def get_yarn_cmd(self, plugin):
        return os.path.join(plugin._npm_dir, "bin", "yarn")

    def test_pull_with_npm(self):
        self.options.nodejs_package_manager = "npm"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        plugin.pull()

        expected_tar_calls = [
            mock.call(self.nodejs_url, plugin._npm_dir),
            mock.call().download(),
            mock.call().provision(
                plugin._npm_dir, clean_target=False, keep_tarball=True
            ),
        ]

        self.tar_mock.assert_has_calls(expected_tar_calls)

    def test_pull_with_yarn(self):
        self.options.nodejs_package_manager = "yarn"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        plugin.pull()

        expected_tar_calls = [
            mock.call(self.nodejs_url, plugin._npm_dir),
            mock.call().download(),
            mock.call("https://yarnpkg.com/latest.tar.gz", plugin._npm_dir),
            mock.call().download(),
            mock.call().provision(
                plugin._npm_dir, clean_target=False, keep_tarball=True
            ),
            mock.call().provision(
                plugin._npm_dir, clean_target=False, keep_tarball=True
            ),
        ]

        self.tar_mock.assert_has_calls(expected_tar_calls)

    def test_build_with_npm(self):
        self.options.nodejs_package_manager = "npm"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        plugin.build()

        self.assertThat(os.path.join(plugin.installdir, "bin", "run"), FileExists())

        expected_env = dict(PATH=os.path.join(plugin._npm_dir, "bin"))

        expected_run_calls = [
            mock.call(
                [self.get_npm_cmd(plugin), "install", "--unsafe-perm"],
                cwd=plugin.builddir,
                env=expected_env,
            ),
            mock.call(
                [self.get_npm_cmd(plugin), "pack"],
                cwd=plugin.builddir,
                env=expected_env,
            ),
            mock.call(
                [
                    self.get_npm_cmd(plugin),
                    "install",
                    "--unsafe-perm",
                    "--offline",
                    "--prod",
                ],
                cwd=os.path.join(plugin.builddir, "package"),
                env=expected_env,
            ),
        ]
        expected_tar_calls = [
            mock.call("test-nodejs-1.0.tgz", plugin.builddir),
            mock.call().provision(os.path.join(plugin.builddir, "package")),
        ]

        self.run_mock.assert_has_calls(expected_run_calls)
        self.tar_mock.assert_has_calls(expected_tar_calls)

        expected_tar_calls = [
            mock.call("test-nodejs-1.0.tgz", plugin.builddir),
            mock.call().provision(os.path.join(plugin.builddir, "package")),
        ]
        self.tar_mock.assert_has_calls(expected_tar_calls)

    def test_build_with_yarn(self):
        self.options.nodejs_package_manager = "yarn"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        plugin.build()

        self.assertThat(os.path.join(plugin.installdir, "bin", "run"), FileExists())

        expected_env = dict(PATH=os.path.join(plugin._npm_dir, "bin"))
        cmd = [self.get_yarn_cmd(plugin)]
        expected_run_calls = [
            mock.call(cmd + ["install"], cwd=plugin.builddir, env=expected_env),
            mock.call(
                cmd + ["pack", "--filename", "test-nodejs-1.0.tgz"],
                cwd=plugin.builddir,
                env=expected_env,
            ),
            mock.call(
                cmd + ["install", "--offline", "--prod"],
                cwd=os.path.join(plugin.builddir, "package"),
                env=expected_env,
            ),
        ]
        expected_tar_calls = [
            mock.call("test-nodejs-1.0.tgz", plugin.builddir),
            mock.call().provision(os.path.join(plugin.builddir, "package")),
        ]

        self.run_mock.assert_has_calls(expected_run_calls)
        self.tar_mock.assert_has_calls(expected_tar_calls)

        expected_tar_calls = [
            mock.call("test-nodejs-1.0.tgz", plugin.builddir),
            mock.call().provision(os.path.join(plugin.builddir, "package")),
        ]
        self.tar_mock.assert_has_calls(expected_tar_calls)

    def test_build_yarn_with_proxy(self):
        self.useFixture(
            fixtures.EnvironmentVariable("http_proxy", "http://localhost:3132")
        )
        self.useFixture(
            fixtures.EnvironmentVariable("https_proxy", "http://localhost:3133")
        )
        self.options.nodejs_package_manager = "yarn"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        plugin.build()

        self.assertThat(os.path.join(plugin.installdir, "bin", "run"), FileExists())

        cmd = [
            self.get_yarn_cmd(plugin),
            "--proxy",
            "http://localhost:3132",
            "--https-proxy",
            "http://localhost:3133",
        ]

        expected_run_calls = [
            mock.call(cmd + ["install"], cwd=plugin.builddir, env=mock.ANY),
            mock.call(
                cmd + ["pack", "--filename", "test-nodejs-1.0.tgz"],
                cwd=plugin.builddir,
                env=mock.ANY,
            ),
            mock.call(
                cmd + ["install", "--offline", "--prod"],
                cwd=os.path.join(plugin.builddir, "package"),
                env=mock.ANY,
            ),
        ]

        self.run_mock.assert_has_calls(expected_run_calls)

    def test_build_scoped_name_with_npm(self):
        self.options.nodejs_package_manager = "npm"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin, package_name="@org/name")

        plugin.build()

        expected_run_calls = [
            mock.call(
                [self.get_npm_cmd(plugin), "install", "--unsafe-perm"],
                cwd=os.path.join(plugin.builddir),
                env=mock.ANY,
            ),
            mock.call(
                [self.get_npm_cmd(plugin), "pack"], cwd=plugin.builddir, env=mock.ANY
            ),
            mock.call(
                [
                    self.get_npm_cmd(plugin),
                    "install",
                    "--unsafe-perm",
                    "--offline",
                    "--prod",
                ],
                cwd=os.path.join(plugin.builddir, "package"),
                env=mock.ANY,
            ),
        ]

        self.run_mock.assert_has_calls(expected_run_calls)

        expected_tar_calls = [
            mock.call("org-name-1.0.tgz", plugin.builddir),
            mock.call().provision(os.path.join(plugin.builddir, "package")),
        ]
        self.tar_mock.assert_has_calls(expected_tar_calls)

    def test_build_scoped_name_with_yarn(self):
        self.options.nodejs_package_manager = "yarn"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin, package_name="@org/name")

        plugin.build()

        cmd = [self.get_yarn_cmd(plugin)]
        expected_run_calls = [
            mock.call(cmd + ["install"], cwd=plugin.builddir, env=mock.ANY),
            mock.call(
                cmd + ["pack", "--filename", "org-name-1.0.tgz"],
                cwd=plugin.builddir,
                env=mock.ANY,
            ),
            mock.call(
                cmd + ["install", "--offline", "--prod"],
                cwd=os.path.join(plugin.builddir, "package"),
                env=mock.ANY,
            ),
        ]

        self.run_mock.assert_has_calls(expected_run_calls)

        expected_tar_calls = [
            mock.call("org-name-1.0.tgz", plugin.builddir),
            mock.call().provision(os.path.join(plugin.builddir, "package")),
        ]
        self.tar_mock.assert_has_calls(expected_tar_calls)


@pytest.fixture(params=["npm", "yarn"])
def nodejs_plugin(project, request):
    """Return an instance of NodejsPlugin setup varying bases and pkg managers."""

    class Options:
        source = "."
        nodejs_version = nodejs._NODEJS_VERSION
        nodejs_package_manager = request.param
        nodejs_yarn_version = ""
        source = "."

    return nodejs.NodePlugin("test-part", Options(), project)


def _create_assets(nodejs_plugin, mock_tar, package_name="test-nodejs"):
    for directory in (nodejs_plugin.sourcedir, nodejs_plugin.builddir):
        directory_path = pathlib.Path(directory)
        directory_path.mkdir(parents=True)

        package_json_path = directory_path / "package.json"
        with package_json_path.open("w") as json_file:
            json.dump(
                dict(name=package_name, bin=dict(run="node.js"), version="1.0"),
                json_file,
            )

        package_name = package_name.lstrip("@").replace("/", "-")
        packed_tar_path = directory_path / f"{package_name}-1.0.tgz"

        (directory_path / "node.js").touch()

    tarfile.TarFile(packed_tar_path, "w").close()

    def provision(directory, **kwargs):
        if directory == os.path.join(nodejs_plugin.builddir, "package"):
            package_dir = os.path.join(nodejs_plugin.builddir, "package")
            open(os.path.join(package_dir, "node.js"), "w").close()

    # Create a fake node bin
    bin_path = pathlib.Path(nodejs_plugin._npm_dir) / "bin"
    bin_path.mkdir(parents=True)
    (bin_path / "node").touch()

    nodejs_bin_path = pathlib.Path(nodejs_plugin.installdir) / "node.js"
    nodejs_bin_path.parent.mkdir(parents=True)
    nodejs_bin_path.touch()


@pytest.fixture()
def nodejs_plugin_with_assets(mock_tar, nodejs_plugin):
    """Return an instance of NodejsPlugin with artifacts setup using nodejs_plugin."""
    _create_assets(nodejs_plugin, mock_tar)

    return nodejs_plugin


class TestManifest:
    scenarios = [
        (
            "simple",
            dict(
                ls_output=(
                    '{"dependencies": {'
                    '   "testpackage1": {"version": "1.0"},'
                    '   "testpackage2": {"version": "1.2"}}}'
                ),
                expected_dependencies=["testpackage1=1.0", "testpackage2=1.2"],
            ),
        ),
        (
            "nested",
            dict(
                ls_output=(
                    '{"dependencies": {'
                    '   "testpackage1": {'
                    '      "version": "1.0",'
                    '      "dependencies": {'
                    '        "testpackage2": {"version": "1.2"}}}}}'
                ),
                expected_dependencies=["testpackage1=1.0", "testpackage2=1.2"],
            ),
        ),
        (
            "missing",
            dict(
                ls_output=(
                    '{"dependencies": {'
                    '   "testpackage1": {"version": "1.0"},'
                    '   "testpackage2": {"version": "1.2"},'
                    '   "missing": {"noversion": "dummy"}}}'
                ),
                expected_dependencies=["testpackage1=1.0", "testpackage2=1.2"],
            ),
        ),
        ("none", dict(ls_output="{}", expected_dependencies=[])),
    ]

    def test_get_manifest_with_node_packages(
        self,
        mock_run,
        mock_run_output,
        nodejs_plugin_with_assets,
        ls_output,
        expected_dependencies,
    ):
        mock_run_output.return_value = ls_output
        plugin = nodejs_plugin_with_assets

        plugin.build()

        if plugin.options.nodejs_package_manager == "yarn":
            expected_dependencies = []
        assert plugin.get_manifest() == collections.OrderedDict(
            {"node-packages": expected_dependencies}
        )


class NodePluginYarnLockManifestTest(NodePluginBaseTest):
    def test_get_manifest_with_yarn_lock_file(self):
        self.options.nodejs_package_manager = "yarn"
        plugin = nodejs.NodePlugin("test-part", self.options, self.project)

        self.create_assets(plugin)

        with open(os.path.join(plugin.builddir, "yarn.lock"), "w") as yarn_lock_file:
            yarn_lock_file.write("test yarn lock contents")

        plugin.build()

        expected_manifest = collections.OrderedDict()
        expected_manifest["yarn-lock-contents"] = "test yarn lock contents"
        expected_manifest["node-packages"] = []

        self.assertThat(plugin.get_manifest(), Equals(expected_manifest))


class TestNodeBin:
    scenarios = [
        (
            "dict",
            dict(
                package_json=dict(
                    name="package-foo",
                    bin=dict(run1="bin0/run1bin", run2="bin1/run2bin"),
                ),
                expected_bins=["run1", "run2"],
            ),
        ),
        (
            "single",
            dict(
                package_json=dict(name="package-foo", bin="bin0/run1bin"),
                expected_bins=["package-foo"],
            ),
        ),
        (
            "single, scoped package",
            dict(
                package_json=dict(name="@org/package-foo", bin="bin1/run1bin"),
                expected_bins=["package-foo"],
            ),
        ),
    ]

    def test_bins(self, tmp_work_path, package_json, expected_bins):
        if type(package_json["bin"]) == dict:
            binaries = package_json["bin"].values()
        else:
            binaries = [package_json["bin"]]

        for binary in binaries:
            os.makedirs(os.path.dirname(binary), exist_ok=True)
            open(binary, "w").close()

        nodejs._create_bins(package_json, ".")
        binary_paths = (os.path.join("bin", b) for b in expected_bins)

        for binary in binary_paths:
            assert os.path.exists(binary)


class TestNodeRelease:
    scenarios = [
        (
            "amd64",
            dict(
                deb_arch="amd64",
                engine="4.4.4",
                expected_url=(
                    "https://nodejs.org/dist/v4.4.4/node-v4.4.4-linux-x64.tar.gz"
                ),
            ),
        ),
        (
            "i386",
            dict(
                deb_arch="i386",
                engine="4.4.4",
                expected_url=(
                    "https://nodejs.org/dist/v4.4.4/node-v4.4.4-linux-x86.tar.gz"
                ),
            ),
        ),
        (
            "armhf",
            dict(
                deb_arch="armhf",
                engine="4.4.4",
                expected_url=(
                    "https://nodejs.org/dist/v4.4.4/node-v4.4.4-linux-armv7l.tar.gz"
                ),
            ),
        ),
        (
            "aarch64",
            dict(
                deb_arch="arm64",
                engine="4.4.4",
                expected_url=(
                    "https://nodejs.org/dist/v4.4.4/node-v4.4.4-linux-arm64.tar.gz"
                ),
            ),
        ),
        (
            "s390x",
            dict(
                deb_arch="s390x",
                engine="4.4.4",
                expected_url=(
                    "https://nodejs.org/dist/v4.4.4/node-v4.4.4-linux-s390x.tar.gz"
                ),
            ),
        ),
    ]

    def test_get_nodejs_release(self, deb_arch, engine, expected_url):
        assert nodejs.get_nodejs_release(engine, deb_arch) == expected_url


class NodePluginUnsupportedArchTest(NodePluginBaseTest):
    @mock.patch("snapcraft.project.Project.deb_arch", "ppcel64")
    def test_unsupported_arch_raises_exception(self):
        raised = self.assertRaises(
            errors.SnapcraftEnvironmentError,
            nodejs.NodePlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(
            raised.__str__(), Equals("architecture not supported (ppcel64)")
        )


def test_missing_package_json(mock_run, mock_run_output, nodejs_plugin_with_assets):
    (pathlib.Path(nodejs_plugin_with_assets.sourcedir) / "package.json").unlink()

    with pytest.raises(nodejs.NodejsPluginMissingPackageJsonError):
        nodejs_plugin_with_assets.pull()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Contains, Equals

from . import FakeStoreCommandsBaseTestCase


class PromoteCommandTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_store_status.mock.return_value = {
            "channel_map_tree": {
                "latest": {
                    "16": {
                        "amd64": [
                            {"channel": "stable", "info": "none"},
                            {
                                "channel": "candidate",
                                "info": "specific",
                                "revision": 1,
                                "version": "0.1",
                            },
                            {
                                "channel": "beta",
                                "info": "specific",
                                "revision": 2,
                                "version": "0.1",
                            },
                            {
                                "channel": "edge",
                                "info": "specific",
                                "revision": 5,
                                "version": "0.1",
                            },
                        ]
                    }
                }
            }
        }
        self.fake_store_release.mock.return_value = {"opened_channels": ["candidate"]}

    def test_upload_without_snap_must_raise_exception(self):
        result = self.run_command(["promote"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_promote_edge_with_yes(self):
        result = self.run_command(
            [
                "promote",
                "test-snap",
                "--from-channel",
                "edge",
                "--to-channel",
                "candidate",
                "--yes",
            ]
        )

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(
            result.output,
            Contains(
                "'edge' is not a valid set value for --from-channel when using --yes."
            ),
        )

    def test_promote_same(self):
        result = self.run_command(
            [
                "promote",
                "test-snap",
                "--from-channel",
                "candidate",
                "--to-channel",
                "candidate",
            ]
        )

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(
            result.output,
            Contains("--from-channel and --to-channel cannot be the same."),
        )

    def test_promote_confirm_yes(self):
        result = self.run_command(
            [
                "promote",
                "snap-test",
                "--from-channel",
                "beta",
                "--to-channel",
                "candidate",
            ],
            input="y\n",
        )

        self.assertThat(result.exit_code, Equals(0))
        self.fake_store_release.mock.assert_called_once_with(
            snap_name="snap-test", revision="2", channels=["candidate"]
        )

    def test_promote_yes_option(self):
        result = self.run_command(
            [
                "promote",
                "snap-test",
                "--from-channel",
                "beta",
                "--to-channel",
                "candidate",
                "--yes",
            ]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.fake_store_release.mock.assert_called_once_with(
            snap_name="snap-test", revision="2", channels=["candidate"]
        )

    def test_promote_confirm_no(self):
        result = self.run_command(
            [
                "promote",
                "snap-test",
                "--from-channel",
                "beta",
                "--to-channel",
                "candidate",
            ],
            input="n\n",
        )

        self.assertThat(result.exit_code, Equals(0))
        self.fake_store_release.mock.assert_not_called()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Contains, Equals

from snapcraft import storeapi
from snapcraft.storeapi.v2.channel_map import (
    MappedChannel,
    Progressive,
    Revision,
    SnapChannel,
)

from . import FakeStoreCommandsBaseTestCase


class ReleaseCommandTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_store_release.mock.return_value = {"opened_channels": ["2.1/beta"]}

    def test_release_without_snap_name_must_raise_exception(self):
        result = self.run_command(["release"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_release(self):
        result = self.run_command(["release", "nil-snap", "19", "2.1/beta"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      amd64   stable     -          -
                             candidate  -          -
                             beta       10         19
                             edge       ↑          ↑
            The '2.1/beta' channel is now open.
            """
                )
            ),
        )
        self.fake_store_release.mock.assert_called_once_with(
            snap_name="nil-snap",
            revision="19",
            channels=["2.1/beta"],
            progressive_percentage=None,
        )

    def test_progressive_release(self):
        self.channel_map.channel_map[0].progressive.percentage = 10.0
        self.channel_map.channel_map[0].progressive.current_percentage = 5.0

        result = self.run_command(
            [
                "release",
                "nil-snap",
                "19",
                "2.1/beta",
                "--progressive",
                "10",
                "--experimental-progressive-releases",
            ]
        )

        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            *EXPERIMENTAL* progressive releases in use.
            Track    Arch    Channel    Version    Revision    Progress
            2.1      amd64   stable     -          -           -
                             candidate  -          -           -
                             beta       -          -           -
                                        10         19          5→10%
                             edge       ↑          ↑           -
            The '2.1/beta' channel is now open.
            """
                )
            ),
        )
        self.fake_store_release.mock.assert_called_once_with(
            snap_name="nil-snap",
            revision="19",
            channels=["2.1/beta"],
            progressive_percentage=10,
        )

    def test_release_with_branch(self):
        self.fake_store_release.mock.return_value = {
            "opened_channels": ["stable/hotfix1"]
        }
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/stable/hotfix1",
                architecture="amd64",
                expiration_date="2020-02-03T20:58:37Z",
                revision=20,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["amd64"], revision=20, version="10hotfix")
        )
        self.channel_map.snap.channels.append(
            SnapChannel(
                name="2.1/stable/hotfix1",
                track="2.1",
                risk="stable",
                branch="hotfix1",
                fallback="2.1/stable",
            )
        )

        result = self.run_command(["release", "nil-snap", "20", "2.1/stable/hotfix1"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel         Version    Revision    Expires at
            2.1      amd64   stable          -          -
                             stable/hotfix1  10hotfix   20          2020-02-03T20:58:37Z
                             candidate       -          -
                             beta            10         19
                             edge            ↑          ↑
            The 'stable/hotfix1' channel is now open.
                    """
                )
            ),
        )

        self.fake_store_release.mock.assert_called_once_with(
            snap_name="nil-snap",
            revision="20",
            channels=["2.1/stable/hotfix1"],
            progressive_percentage=None,
        )

    def test_progressive_release_with_branch(self):
        self.fake_store_release.mock.return_value = {
            "opened_channels": ["2.1/stable/hotfix1"]
        }
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/stable/hotfix1",
                architecture="amd64",
                expiration_date="2020-02-03T20:58:37Z",
                revision=20,
                progressive=Progressive(
                    paused=None, percentage=80.0, current_percentage=None
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["amd64"], revision=20, version="10hotfix")
        )
        self.channel_map.snap.channels.append(
            SnapChannel(
                name="2.1/stable/hotfix1",
                track="2.1",
                risk="stable",
                branch="hotfix1",
                fallback="2.1/stable",
            )
        )

        result = self.run_command(
            [
                "release",
                "--progressive",
                "80",
                "--experimental-progressive-releases",
                "nil-snap",
                "20",
                "2.1/stable/hotfix1",
            ]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            *EXPERIMENTAL* progressive releases in use.
            Track    Arch    Channel         Version    Revision    Progress    Expires at
            2.1      amd64   stable          -          -           -
                             stable/hotfix1  10hotfix   20          ?→80%       2020-02-03T20:58:37Z
                             candidate       -          -           -
                             beta            10         19          -
                             edge            ↑          ↑           -
            The '2.1/stable/hotfix1' channel is now open.
                    """
                )
            ),
        )
        self.fake_store_release.mock.assert_called_once_with(
            snap_name="nil-snap",
            revision="20",
            channels=["2.1/stable/hotfix1"],
            progressive_percentage=80,
        )

    def test_progressive_release_with_null_current_percentage(self):
        self.channel_map.channel_map[0].progressive.percentage = 10.0
        self.channel_map.channel_map[0].progressive.current_percentage = None

        result = self.run_command(
            [
                "release",
                "nil-snap",
                "19",
                "2.1/beta",
                "--progressive",
                "10",
                "--experimental-progressive-releases",
            ]
        )

        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            *EXPERIMENTAL* progressive releases in use.
            Track    Arch    Channel    Version    Revision    Progress
            2.1      amd64   stable     -          -           -
                             candidate  -          -           -
                             beta       -          -           -
                                        10         19          ?→10%
                             edge       ↑          ↑           -
            The '2.1/beta' channel is now open.
            """
                )
            ),
        )
        self.fake_store_release.mock.assert_called_once_with(
            snap_name="nil-snap",
            revision="19",
            channels=["2.1/beta"],
            progressive_percentage=10,
        )

    def test_release_without_login_must_ask(self):
        self.fake_store_release.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            {"opened_channels": ["beta"]},
        ]

        result = self.run_command(
            ["release", "nil-snap", "19", "beta"], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from jsonschema.exceptions import ValidationError

from snapcraft.storeapi.v2 import whoami


@pytest.fixture
def whoami_account_payload():
    return {
        "email": "foo@bar.baz",
        "id": "1234567890",
        "name": "Foo from Bar",
        "username": "foo",
    }


@pytest.fixture
def whoami_payload(whoami_account_payload):
    return {
        "account": whoami_account_payload,
        "channels": None,
        "packages": None,
        "permissions": None,
    }


def test_account(whoami_account_payload):
    w = whoami.Account.unmarshal(whoami_account_payload)

    assert repr(w) == "<Account: 'foo@bar.baz'>"
    assert w.email == whoami_account_payload["email"]
    assert w.account_id == whoami_account_payload["id"]
    assert w.name == whoami_account_payload["name"]
    assert w.username == whoami_account_payload["username"]
    assert w.marshal() == whoami_account_payload


@pytest.mark.parametrize("missing", ("email", "id", "name", "username"))
def test_account_missing_data(missing, whoami_account_payload):
    whoami_account_payload.pop(missing)

    with pytest.raises(ValidationError):
        whoami.Account.unmarshal(whoami_account_payload)


def test_whoami(whoami_payload):
    a = whoami.WhoAmI.unmarshal(whoami_payload)

    assert repr(a) == "<WhoAmI: 'foo@bar.baz'>"
    assert a.account.email == whoami_payload["account"]["email"]
    assert a.account.account_id == whoami_payload["account"]["id"]
    assert a.account.name == whoami_payload["account"]["name"]
    assert a.account.username == whoami_payload["account"]["username"]
    # TODO: implement fully to migrate away from account_info.
    assert a.marshal()["account"] == whoami_payload["account"]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import qmake

from . import PluginsV1BaseTestCase


class QMakeTestCase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.internal.common.run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("sys.stdout")
        patcher.start()
        self.addCleanup(patcher.stop)

        class Options:
            options = []
            qt_version = "qt5"
            project_files = []

        self.options = Options()

    def test_schema(self):
        schema = qmake.QmakePlugin.schema()

        # Verify the presence of all properties
        properties = schema["properties"]
        self.assertTrue(
            "options" in properties, 'Expected "options" to be included in properties'
        )
        self.assertTrue(
            "qt-version" in properties,
            'Expected "qt-version" to be included in properties',
        )

        # Check options property
        options = properties["options"]
        for item in ["type", "minitems", "uniqueItems", "items", "default"]:
            self.assertTrue(
                item in options,
                'Expected {!r} to be included in "options"'.format(item),
            )

        options_type = options["type"]
        self.assertThat(
            options_type,
            Equals("array"),
            'Expected "options" "type" to be "array", but it '
            'was "{}"'.format(options_type),
        )

        options_minitems = options["minitems"]
        self.assertThat(
            options_minitems,
            Equals(1),
            'Expected "options" "minitems" to be 1, but '
            "it was {}".format(options_minitems),
        )

        self.assertTrue(options["uniqueItems"])

        options_default = options["default"]
        self.assertThat(
            options_default,
            Equals([]),
            'Expected "options" "default" to be [], but '
            "it was {}".format(options_default),
        )

        options_items = options["items"]
        self.assertTrue(
            "type" in options_items,
            'Expected "type" to be included in "options" ' '"items"',
        )

        options_items_type = options_items["type"]
        self.assertThat(
            options_items_type,
            Equals("string"),
            'Expected "options" "items" "type" to be '
            '"string", but it was "{}"'.format(options_items_type),
        )

        # Check qt-version property
        qt_version = properties["qt-version"]
        self.assertTrue(
            "enum" in qt_version, 'Expected "enum" to be included in "qt_version"'
        )

        qt_version_enum = qt_version["enum"]
        # Using sets for order independence in the comparison
        self.assertThat(set(qt_version_enum), Equals(set(["qt4", "qt5"])))

        qt_version_default = qt_version["default"]
        self.assertThat(
            qt_version_default,
            Equals("qt5"),
            'Expected "qt_version" "default" to be "qt5", but '
            "it was {}".format(qt_version_default),
        )

        # Check project-files property
        project_files = properties["project-files"]
        for item in ["type", "minitems", "uniqueItems", "items", "default"]:
            self.assertTrue(
                item in options,
                'Expected {!r} to be included in "project_files"'.format(item),
            )

        project_files_type = project_files["type"]
        self.assertThat(
            project_files_type,
            Equals("array"),
            'Expected "project_files" "type" to be "array", but '
            'it was "{}"'.format(project_files_type),
        )

        project_files_minitems = project_files["minitems"]
        self.assertThat(
            project_files_minitems,
            Equals(1),
            'Expected "project_files" "minitems" to be 1, but '
            "it was {}".format(project_files_minitems),
        )

        self.assertTrue(project_files["uniqueItems"])

        project_files_default = project_files["default"]
        self.assertThat(
            project_files_default,
            Equals([]),
            'Expected "project_files" "default" to be [], but '
            "it was {}".format(project_files_default),
        )

        project_files_items = project_files["items"]
        self.assertTrue(
            "type" in project_files_items,
            'Expected "type" to be included in "project_files" ' '"items"',
        )

        project_files_items_type = project_files_items["type"]
        self.assertThat(
            project_files_items_type,
            Equals("string"),
            'Expected "project_files" "items" "type" to be '
            '"string", but it was "{}"'.format(project_files_items_type),
        )

    def test_get_build_properties(self):
        expected_build_properties = ["options", "project-files"]
        resulting_build_properties = qmake.QmakePlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)

    def test_get_pull_properties(self):
        expected_pull_properties = ["qt-version"]
        resulting_pull_properties = qmake.QmakePlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_unsupported_qt_version(self):
        self.options.qt_version = "qt3"
        raised = self.assertRaises(
            RuntimeError, qmake.QmakePlugin, "test-part", self.options, self.project
        )

        self.assertThat(str(raised), Equals("Unsupported Qt version: 'qt3'"))

    def test_build_referencing_sourcedir_if_no_subdir(self):
        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(["qmake"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_referencing_sourcedir_with_subdir(self):
        self.options.source_subdir = "subdir"

        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(["qmake"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_with_project_file(self):
        self.options.project_files = ["project_file.pro"]

        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        path_to_pro_file = os.path.join(plugin.sourcedir, "project_file.pro")
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["qmake", path_to_pro_file], cwd=plugin.builddir, env=mock.ANY
                ),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_with_project_file_with_subdir(self):
        self.options.source_subdir = "subdir"
        self.options.project_files = ["project_file.pro"]

        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        path_to_pro_file = os.path.join(
            plugin.sourcedir, plugin.options.source_subdir, "project_file.pro"
        )
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["qmake", path_to_pro_file], cwd=plugin.builddir, env=mock.ANY
                ),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_with_options(self):
        self.options.options = ["-foo"]

        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(["qmake", "-foo"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_with_libs_and_includes_in_installdir(self):
        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        os.makedirs(os.path.join(plugin.installdir, "include"))
        os.makedirs(os.path.join(plugin.installdir, "lib"))
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        "qmake",
                        'LIBS+="-L{}/lib"'.format(plugin.installdir),
                        'INCLUDEPATH+="{}/include"'.format(plugin.installdir),
                    ],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_with_libs_and_includes_in_stagedir(self):
        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        os.makedirs(os.path.join(plugin.project.stage_dir, "include"))
        os.makedirs(os.path.join(plugin.project.stage_dir, "lib"))
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    [
                        "qmake",
                        'LIBS+="-L{}/lib"'.format(plugin.project.stage_dir),
                        'INCLUDEPATH+="{}/include"'.format(plugin.project.stage_dir),
                    ],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(["make", "-j2"], cwd=plugin.builddir, env=mock.ANY),
                mock.call(
                    ["make", "install", "INSTALL_ROOT={}".format(plugin.installdir)],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            qmake.QmakePlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))

    def test_build_environment_qt4(self):
        self.options.qt_version = "qt4"
        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        self.assert_expected_environment({"QT_SELECT": "qt4"})

    def test_build_environment_qt5(self):
        self.options.qt_version = "qt5"
        plugin = qmake.QmakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)
        plugin.build()

        self.assert_expected_environment({"QT_SELECT": "qt5"})

    def assert_expected_environment(self, expected):
        self.assertThat(self.run_mock.call_count, Equals(3))
        for call_args in self.run_mock.call_args_list:
            environment = call_args[1]["env"]
            for variable, value in expected.items():
                self.assertTrue(
                    variable in environment,
                    "Expected variable {!r} to be in environment".format(variable),
                )

                self.assertThat(
                    environment[variable],
                    Equals(value),
                    "Expected ${}={}, but it was {}".format(
                        variable, value, environment[variable]
                    ),
                )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os.path
import shutil
import subprocess
import sys
import tempfile

import requests
import simplejson

from snapcraft.file_utils import calculate_sha3_384
from snapcraft.internal.build_providers.errors import (
    ProviderMultipassDownloadFailed,
    ProviderMultipassInstallationFailed,
)
from snapcraft.internal.indicators import download_requests_stream

if sys.platform == "win32":
    import winreg


logger = logging.getLogger(__name__)


_MULTIPASS_RELEASES_API_URL = (
    "https://api.github.com/repos/CanonicalLtd/multipass/releases/latest"
)
_MULTIPASS_DL_VERSION = "0.8.0"
_MULTIPASS_DL_NAME = "multipass-{version}+win-win64.exe".format(
    version=_MULTIPASS_DL_VERSION
)
_MULTIPASS_DL_SHA3_384 = "a1ac2eeb77b2a98fe5dee198be70dbf1a985d94b9707ce33ea0d3828dbc90d07fccb9662b7c97a3cfa194895b4f56676"


def windows_reload_multipass_path_env():
    """Update PATH to include installed Multipass, if not already set."""

    assert sys.platform == "win32"

    key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Environment")

    paths = os.environ["PATH"].split(";")

    # Drop empty placeholder for trailing comma, if present.
    if paths[-1] == "":
        del paths[-1]

    reg_user_path, _ = winreg.QueryValueEx(key, "Path")
    for path in reg_user_path.split(";"):
        if path not in paths and "Multipass" in path:
            paths.append(path)

    # Restore path with trailing comma.
    os.environ["PATH"] = ";".join(paths) + ";"


def _run_installer(installer_path: str, echoer):
    """Execute multipass installer."""

    echoer.info("Installing Multipass...")

    # Multipass requires administrative privileges to install, which requires
    # the use of `runas` functionality. Some of the options included:
    # (1) https://stackoverflow.com/a/34216774
    # (2) ShellExecuteW and wait on installer by attempting to delete it.
    # Windows would prevent us from deleting installer with a PermissionError:
    # PermissionError: [WinError 32] The process cannot access the file because
    # it is being used by another process: <path>
    # (3) Use PowerShell's "Start-Process" with RunAs verb as shown below.
    # None of the options are quite ideal, but #3 will do.
    cmd = """
    & {{
        try {{
            $Output = Start-Process -FilePath {path!r} -Args /S -Verb RunAs -Wait -PassThru
        }} catch {{
            [Environment]::Exit(1)
        }}
      }}
    """.format(
        path=installer_path
    )

    try:
        subprocess.check_call(["powershell.exe", "-Command", cmd])
    except subprocess.CalledProcessError:
        raise ProviderMultipassInstallationFailed("error launching installer")

    # Reload path environment to see if we can find multipass now.
    windows_reload_multipass_path_env()

    if not shutil.which("multipass.exe"):
        # Installation failed.
        raise ProviderMultipassInstallationFailed(
            "installation did not complete successfully"
        )

    echoer.info("Multipass installation completed successfully.")


def _requests_exception_hint(e: requests.RequestException) -> str:
    # Use the __doc__ description to give the user a hint. It seems to be a
    # a decent option over trying to enumerate all of possible types.
    if e.__doc__:
        split_lines = e.__doc__.splitlines()
        if split_lines:
            return e.__doc__.splitlines()[0].strip()

    # Should never get here.
    return "unknown download error"


def _fetch_installer_url() -> str:
    """Fetch latest installer executable from github.

    If the latest release on github is newer than what snapcraft knows
    about in _MULTIPASS_DL_NAME_*, we skip the download and inform the
    user to go download it manually.  This way we will only ever directly
    execute whitelisted executables on behalf of the user.  Verify the
    installer using a SHA3-384 digest.
    """

    try:
        resp = requests.get(_MULTIPASS_RELEASES_API_URL)
    except requests.RequestException as e:
        raise ProviderMultipassDownloadFailed(_requests_exception_hint(e))

    try:
        data = resp.json()
    except simplejson.scanner.JSONDecodeError:
        raise ProviderMultipassDownloadFailed(
            "failed to fetch valid release data from {}".format(
                _MULTIPASS_RELEASES_API_URL
            )
        )

    for asset in data.get("assets", list()):
        # Find matching name.
        if asset.get("name") != _MULTIPASS_DL_NAME:
            continue

        return asset.get("browser_download_url")

    # Something changed we don't know about - we will simply categorize
    # all possible events as an updated version we do not yet know about.
    raise ProviderMultipassDownloadFailed(
        "an updated version is available that snapcraft does not yet know about"
    )


def _download_multipass(dl_dir: str, echoer) -> str:
    """Creates temporary Downloads installer to temp directory."""

    dl_url = _fetch_installer_url()
    dl_basename = os.path.basename(dl_url)
    dl_path = os.path.join(dl_dir, dl_basename)

    echoer.info("Downloading Multipass installer...\n{} -> {}".format(dl_url, dl_path))

    try:
        request = requests.get(dl_url, stream=True, allow_redirects=True)
        request.raise_for_status()
        download_requests_stream(request, dl_path)
    except requests.RequestException as e:
        raise ProviderMultipassDownloadFailed(_requests_exception_hint(e))

    digest = calculate_sha3_384(dl_path)
    if digest != _MULTIPASS_DL_SHA3_384:
        raise ProviderMultipassDownloadFailed(
            "download failed verification (expected={} but found={})".format(
                _MULTIPASS_DL_SHA3_384, digest
            )
        )

    echoer.info("Verified installer successfully...")
    return dl_path


def windows_install_multipass(echoer) -> None:
    """Download and install multipass."""

    assert sys.platform == "win32"

    dl_dir = tempfile.mkdtemp()
    dl_path = _download_multipass(dl_dir, echoer)
    _run_installer(dl_path, echoer)

    # Cleanup.
    shutil.rmtree(dl_dir)
